local _G = getfenv(0)
-- Local versions for performance
local string = _G.string
local math = _G.math
local pairs = _G.pairs
local ipairs = _G.ipairs
local table = _G.table
local floor, ceil, min, max, abs = math.floor, math.ceil, math.min, math.max, math.abs
local tconcat, tinsert = table.concat, table.insert
local wipe = _G.wipe
local tostring = _G.tostring
local tonumber = _G.tonumber

local DKDiseases = _G.LibStub("AceAddon-3.0"):NewAddon("DKDiseases", "AceConsole-3.0", "AceEvent-3.0","AceTimer-3.0")
local L = _G.LibStub("AceLocale-3.0"):GetLocale("DKDiseases", true)
local LDB = _G.LibStub("LibDataBroker-1.1")
local LibQTip = _G.LibStub("LibQTip-1.0")
local Icon = _G.LibStub("LibDBIcon-1.0")
local LSM = _G.LibStub:GetLibrary("LibSharedMedia-3.0")

-- Try to remove the Git hash at the end, otherwise return the passed in value.
local function cleanupVersion(version)
	local iter = string.gmatch(version, "(.*)-[a-z0-9]+$")
	if iter then
		local ver = iter()
		if ver and #ver >= 3 then
			return ver
		end
	end
	return version
end

local ADDON_NAME = ...
local ADDON_VERSION = cleanupVersion("0.1-beta-2-gc55b6d9")
local options = nil
local MIN_UPDATE_TIME = 0.05

local function round(number)
    if not number then return 0 end
    return floor(number+0.5)
end

local HEXFMT = "|c%2x%2x%2x%2x"
local function chatColorCode(color)
	local a = 255
	local r = ceil(255 * color.r)
	local g = ceil(255 * color.g)
	local b = ceil(255 * color.b)
	return HEXFMT:format(a, r, g, b)
end

-- Define Tracker for now but the rest is at the bottom of the file.
local Tracker = {}

DKDiseases.trackers = {}
DKDiseases.playerName = UnitName("player")
DKDiseases.isDK = false
DKDiseases.currentSpec = ""
DKDiseases.currentTarget = nil
DKDiseases.targetHostile = false
DKDiseases.baseFF = 0
DKDiseases.baseBP = 0
DKDiseases.effFF = 0
DKDiseases.effBP = 0
DKDiseases.FF = {}
DKDiseases.BP = {}
DKDiseases.cinderglacier = false
DKDiseases.effectiveAP = 0
DKDiseases.mastery = 0
DKDiseases.critChance = 0
DKDiseases.damageModifier = 1.0

local Broker = _G.CreateFrame("Frame")
Broker.obj = LDB:NewDataObject(_G.GetAddOnMetadata(ADDON_NAME, "Title"), {
    type = "data source",
    icon = "Interface\\Icons\\Spell_DeathKnight_BloodPlague",
    label = _G.GetAddOnMetadata(ADDON_NAME, "Title"),
    text = _G.GetAddOnMetadata(ADDON_NAME, "Title"),
    barValue = 0,
    barR = 0,
    barG = 0,
    barB = 1,
	OnClick = function(clickedframe, button)
		if button == "RightButton" then
			local optionsFrame = _G.InterfaceOptionsFrame

			if optionsFrame:IsVisible() then
				optionsFrame:Hide()
			else
				DKDiseases:ShowOptions()
			end
		elseif button == "LeftButton" and _G.IsShiftKeyDown() then
        end
	end
} )

local GREEN = "|cff00ff00"
local YELLOW = "|cffffff00"
local BLUE = "|cff0198e1"
local ORANGE = "|cffff9933"
local addonHdr = GREEN.."%s %s"

function Broker.obj:OnEnter()
	local tooltip = LibQTip:Acquire("DKDiseasesTooltip", 2, "LEFT", "RIGHT")
	self.tooltip = tooltip 

    tooltip:AddHeader(addonHdr:format(_G.GetAddOnMetadata(ADDON_NAME,"Title"), ADDON_VERSION))
    tooltip:AddLine()

    if DKDiseases.isDK then
    end

	tooltip:SmartAnchorTo(self)
	tooltip:Show()
end

function Broker.obj:OnLeave()
	LibQTip:Release(self.tooltip)
	self.tooltip = nil
end

local LookupOrKeyMT = {__index = function (t,k) return k end}
local SpellIds = {
	["Frost Fever"] = 59921,
	["Blood Plague"] = 59879,
	["Frost Fever Periodic"] = 55095,
	["Blood Plague Periodic"] = 55078,
	["Plague Strike"] = 45462,
	["Outbreak"] = 77575,
	["Cinderglacier"] = 53386,
	["Auto Attack"] = 6603,
}
local SpellNames = {}
local SpellIcons = {}
_G.setmetatable(SpellNames, LookupOrKeyMT)
local function LoadSpellNames()
	for k, v in pairs(SpellIds) do
		local name, _, icon = _G.GetSpellInfo(v)
		SpellNames[k] = name
		SpellIcons[k] = icon
	end
end
LoadSpellNames()

local millFmt = "%.0fm"
local thousandFmt = "%.0fk"
local function FormatNumber(number)
    if tonumber(number) == nil then
        return number
    end

    if number > 1000000 then
        return millFmt:format(number / 1000000)
    elseif number > 1000 then
        return thousandFmt:format(number / 1000)
    end

    return number
end

local frames = {}

local defaults = {
	profile = {
		minimap = {
			hide = true,
		},
		debug = false,
		debugBP = false,
		specs = {
			["Frost"] = true,
			["Unholy"] = true,
			["Blood"] = true,
		},
		trackDiseases = true,
		diseaseThresholdType = "Percent",
		diseaseThreshold = 1.1,
		diseaseThresholdAmt = 1000,
		font_face = "Friz Quadrata TT",
		font_size = 13,
		fontFlags = {
			OUTLINE = true,
			THICKOUTLINE = false,
			MONOCHROME = false
		},
		texture = "Blizzard",
		border = "",
		diseaseOrder = "BF",
		reapplycolors = {
			[-1] = { r = 0.8, g = 0.1, b = 0.1, a = 0.5 },
			[0] = { r = 0.0, g = 0.0, b = 0.5, a = 0.8 },
			[1] = { r = 0.8, g = 0.8, b = 0.1, a = 0.5 },
			[2] = { r = 0.1, g = 0.8, b = 0.1, a = 0.5 },
		},
		frames = {
			target = {
				x = 0,
				y = 0,
				inset = 3,
				width = 35,
				height = 35,
				locked = false,
				hideooc = false,
				ooc_alpha = 0.5,
			},
		},
		-- Trackers
		trackers = {
			-- Provide defaults for all trackers.
			-- These are inherited if no bar or no value is set.
			['**'] = {
				textcolor = { r = 1.0, g = 1.0, b = 1.0, a = 1.0 },
				userPlaced = false,
				x = nil,
				y = nil,
			},
			["TargetFrostFever"] = {
				enabled = true,
				color = { r = 0.0, g = 0.45, b = 0.6, a = 1.0 },
				bgcolor = { r = 0.0, g = 0.1125, b = 0.15, a = 0.9 },
				textcolor = { r = 0.8, g = 0.8, b = 0.8, a = 1.0 },
			},
			["TargetBloodPlague"] = {
				enabled = true,
				color = { r = 0.5, g = 0.0, b = 0.6, a = 1.0 },
				bgcolor = { r = 0.125, g = 0.0, b = 0.15, a = 0.9 },
				textcolor = { r = 0.8, g = 0.8, b = 0.8, a = 1.0 },
			},
		},
	},
}

function DKDiseases:GetOptions()
    if not options then
        options = {
            type = "group",
            name = _G.GetAddOnMetadata(ADDON_NAME, "Title"),
            args = {
				core = self:GetCoreOptions(),
            }
        }
		options.args.profile = _G.LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
    end
    return options
end

function DKDiseases:ShowOptions()
	_G.InterfaceOptionsFrame_OpenToCategory(self.optionsFrame.Main)
end

function DKDiseases:GetCoreOptions()
	local core = {
	    order = 1,
		name = L["General Options"],
		type = "group",
		inline = true,
		args = {
		    description = {
		        order = 1,
		        type = "description",
		        name = L["Addon_Desc"],
		    },
		    generalOptions = {
		        order = 100,
		        type = "group",
		        name = L["General Options"],
				args = self:GetGeneralOptions(),
		    },
			diseaseTracking = {
		        order = 200,
		        type = "group",
		        name = L["Disease Strength"],
				args = self:GetStrengthOptions(),
			},
		    dimensions = {
		        order = 500,
		        type = "group",
		        name = L["Dimensions"],
				args = {
					width = {
						order = 510,
						name = L["Width"],
						desc = L["BarWidth_Desc"],	
						type = "range",
						min = 20,
						max = 100,
						step = 1,
						set = function(info, val)
						    self.db.profile.frames.target.width = val
							if frames.target then frames.target:Resize() end
						end,
						get = function(info, val)
						    return self.db.profile.frames.target.width
						end,
					},
					height = {
						order = 520,
						name = L["Height"],
						desc = L["BarHeight_Desc"],
						type = "range",
						min = 5,
						max = 50,
						step = 1,
						set = function(info, val)
						    self.db.profile.frames.target.height = val
							if frames.target then frames.target:Resize() end
						end,
						get = function(info, val)
						    return self.db.profile.frames.target.height
						end,					
					},
				},
		    },
		    fonts = {
		        order = 600,
		        type = "group",
		        name = L["Font"],
				args = {
					font_size = {
						order = 610,
						name = L["Font size"],
						desc = L["Font size for the bars."],
						type = "range",
						min = 8,
						max = 30,
						step = 1,
						set = function(info, val) 
							self.db.profile.font_size = val
							self:ResetFonts()
						end,
						get = function(info,val) return self.db.profile.font_size end,
					},
					font_face = {
						order = 620,
						type = "select",
						name = L["Font"],
						desc = L["Font to use."],
						values = LSM:HashTable("font"),
						dialogControl = 'LSM30_Font',
						get = function() return self.db.profile.font_face end,
						set = function(info, val) 
						    self.db.profile.font_face = val
							self:ResetFonts()
						end
					},
					font_outline = {
						name = L["Outline"],
						desc = L["FontOutlineDesc"],
						type = "toggle",
						order = 630,
						set = function(info, val)
						    self.db.profile.fontFlags.OUTLINE = val
						    self:ResetFonts()
						end,
		                get = function(info)
		                    return self.db.profile.fontFlags.OUTLINE
		                end,
					},
					font_thickoutline = {
						name = L["Thick Outline"],
						desc = L["FontThickOutline_OptionDesc"],
						type = "toggle",
						order = 640,
						set = function(info, val)
						    self.db.profile.fontFlags.THICKOUTLINE = val
						    self:ResetFonts()
						end,
		                get = function(info)
		                    return self.db.profile.fontFlags.THICKOUTLINE
		                end,
					},
					font_monochrome = {
						name = L["Monochrome"],
						desc = L["FontMonochromeDesc"],
						type = "toggle",
						order = 650,
						set = function(info, val)
						    self.db.profile.fontFlags.MONOCHROME = val
						    self:ResetFonts()
						end,
		                get = function(info)
		                    return self.db.profile.fontFlags.MONOCHROME
		                end,
					},
				},
		    },
            textures = {
                order = 700,
                type = "group",
                name = L["Textures"],
				args = {
					bartexture = {
						order = 110,
						name = L["Bar Texture"],
						desc = L["BarTexture_OptionDesc"],
						type = "select",
						values = LSM:HashTable("statusbar"),
						dialogControl = 'LSM30_Statusbar',
						get = function()
						    return self.db.profile.texture
						end,
						set = function(info, val)
						    self.db.profile.texture = val
							self:UpdateTextures()
						end,
					},
					bordertexture = {
						order = 120,
						name = L["Border"],
						desc = L["Border_OptionDesc"],
						type = "select",
						values = LSM:HashTable("border"),
						dialogControl = 'LSM30_Border',
						get = function()
						    return self.db.profile.border
						end,
						set = function(info, val)
						    self.db.profile.border = val
						end,
					},
				},
            },
		},
	}
	return core
end

function DKDiseases:GetGeneralOptions()
	local options = {
		locked = {
			name = L["Lock"],
			desc = L["LockDesc"],
			type = "toggle",
			order = 10,
			set = function(info, val)
			    self.db.profile.frames.target.locked = val
				if frames.target then
					frames.target.lock = val
					frames.target:EnableMouse(not val)
				end
			end,
            get = function(info)
				return self.db.profile.frames.target.locked
			end,
		},
		hideooc = {
			name = L["Hide Out of Combat"],
			desc = L["Hide Out of Combat"],
			type = "toggle",
			order = 15,
			set = function(info, val)
			    self.db.profile.frames.target.hideooc = val
				if frames.target and not _G.UnitAffectingCombat("player") then
					if val then
						frames.target:Hide()
					else
						if self.isDK and self.db.profile.specs[self.currentSpec] then
							frames.target:Show()
						end
					end
				end
			end,
            get = function(info)
				return self.db.profile.frames.target.hideooc
			end,
		},
	    minimap = {
			order = 20,
            name = L["Minimap Button"],
            desc = L["Toggle the minimap button"],
            type = "toggle",
            set = function(info,val)
                	-- Reverse the value since the stored value is to hide it
                    self.db.profile.minimap.hide = not val
                	if self.db.profile.minimap.hide then
                		Icon:Hide("DKDiseasesLDB")
                	else
                		Icon:Show("DKDiseasesLDB")
                	end
                  end,
            get = function(info)
        	        -- Reverse the value since the stored value is to hide it
                    return not self.db.profile.minimap.hide
                  end,
        },
		diseaseOrder = {
			order = 30,
			name = L["Disease Order"],
			desc = L["DiseaseOrder_Desc"],
			type = "select",
			values = {
			    ["FB"] = SpellNames["Frost Fever"]..", "..SpellNames["Blood Plague"],
			    ["BF"] = SpellNames["Blood Plague"]..", "..SpellNames["Frost Fever"],
			    ["F"] = SpellNames["Frost Fever"],
			    ["B"] = SpellNames["Blood Plague"],
			},
			set = function(info, val)
			    self.db.profile.diseaseOrder = val
				if frames.target.ffBar and frames.target.ffBar.trackerType == "icon" then
					frames.target.ffBlock:UpdateLayout()
					frames.target.bpBlock:UpdateLayout()
				else
				end
			end,
            get = function(info)
                return self.db.profile.diseaseOrder
            end,
		},
		ooc_alpha = {
			order = 40,
			name = L["OOC Alpha"],
			desc = L["OOCAlpha_Desc"],
			type = "range",
			min = 0,
			max = 1,
			step = 0.05,
			set = function(info, val)
			    self.db.profile.frames.target.ooc_alpha = val
				if frames.target and not _G.UnitAffectingCombat("player") then
					frames.target:SetAlpha(self.db.profile.frames.target.ooc_alpha)
				end
			end,
			get = function(info, val)
			    return self.db.profile.frames.target.ooc_alpha
			end,					
		},
		specHdr = {
	        order = 200,
	        type = "header",
	        name = L["Enabled for Specs"],
		},
		specBlood = {
			name = L["Blood"],
			desc = L["Blood"],
			type = "toggle",
			order = 210,
			set = function(info, val)
			    self.db.profile.specs["Blood"] = val
				self:CheckEnabled()
			end,
            get = function(info)
				return self.db.profile.specs["Blood"]
			end,
		},
		specFrost = {
			name = L["Frost"],
			desc = L["Frost"],
			type = "toggle",
			order = 220,
			set = function(info, val)
			    self.db.profile.specs["Frost"] = val
				self:CheckEnabled()
			end,
            get = function(info)
				return self.db.profile.specs["Frost"]
			end,
		},
		specUnholy = {
			name = L["Unholy"],
			desc = L["Unholy"],
			type = "toggle",
			order = 230,
			set = function(info, val)
			    self.db.profile.specs["Unholy"] = val
				self:CheckEnabled()
			end,
            get = function(info)
				return self.db.profile.specs["Unholy"]
			end,
		},
	}
	return options
end

function DKDiseases:GetStrengthOptions()
	local options = {
		trackDiseases = {
			name = L["Enabled"],
			desc = L["Enabled"],
			type = "toggle",
			order = 100,
			set = function(info, val)
			    self.db.profile.trackDiseases = val
				self:ToggleDiseaseTracking()
			end,
            get = function(info)
				return self.db.profile.trackDiseases
			end,
		},
		diseaseThreshold = {
			name = L["Threshold"],
			desc = L["Threshold"],
			type = "select",
			values = {
			    ["Percent"] = L["Percent"],
			    ["Fixed"] = L["Fixed"],
			},
			order = 110,
			set = function(info, val)
			    self.db.profile.diseaseThresholdType = val
			end,
            get = function(info)
                return self.db.profile.diseaseThresholdType
            end,
		},
		diseaseThresholdPerc = {
			order = 120,
			name = L["Percent"],
			desc = L["Percent"],
			type = "range",
			min = 1.0,
			max = 3.0,
			step = 0.01,
			bigStep = 0.05,
			isPercent = true,
			set = function(info, val)
			    self.db.profile.diseaseThreshold = val
			end,
			get = function(info, val)
			    return self.db.profile.diseaseThreshold
			end,
			disabled = function()
				return self.db.profile.diseaseThresholdType ~= "Percent"
			end,
		},
		diseaseThresholdAmt = {
			order = 130,
			name = L["Amount"],
			desc = L["Amount"],
			type = "range",
			min = 0,
			max = 100000,
			step = 1,
			bigStep = 500,
			set = function(info, val)
			    self.db.profile.diseaseThresholdAmt = val
			end,
			get = function(info, val)
			    return self.db.profile.diseaseThresholdAmt
			end,
			disabled = function()
				return self.db.profile.diseaseThresholdType ~= "Fixed"
			end,
		},
		colorsHdr = {
	        order = 200,
	        type = "header",
	        name = L["Colors"],
		},
		reapplyThresh = {
			order = 210,
			name = L["ReapplyThreshold"],
			desc = L["ReapplyThreshold_Desc"],
			type = "color",
			hasAlpha = true,
			width = "full",
			set = function(info, r, g, b, a)
			    local c = self.db.profile.reapplycolors[2]
			    c.r, c.g, c.b, c.a = r, g, b, a
			end,
			get = function(info)
		        local c = self.db.profile.reapplycolors[2]
			    return c.r, c.g, c.b, c.a
			end,					
		},
		reapply = {
			order = 220,
			name = L["Reapply"],
			desc = L["Reapply_Desc"],
			type = "color",
			hasAlpha = true,
			width = "full",
			set = function(info, r, g, b, a)
			    local c = self.db.profile.reapplycolors[1]
			    c.r, c.g, c.b, c.a = r, g, b, a
			end,
			get = function(info)
		        local c = self.db.profile.reapplycolors[1]
			    return c.r, c.g, c.b, c.a
			end,					
		},
		noreapply = {
			order = 230,
			name = L["DontReapply"],
			desc = L["DontReapply_Desc"],
			type = "color",
			width = "full",
			hasAlpha = true,
			set = function(info, r, g, b, a)
			    local c = self.db.profile.reapplycolors[-1]
			    c.r, c.g, c.b, c.a = r, g, b, a
			end,
			get = function(info)
		        local c = self.db.profile.reapplycolors[-1]
			    return c.r, c.g, c.b, c.a
			end,					
		},
		nodisease = {
			order = 240,
			name = L["NoDisease"],
			desc = L["NoDisease_Desc"],
			type = "color",
			width = "full",
			hasAlpha = true,
			set = function(info, r, g, b, a)
			    local c = self.db.profile.reapplycolors[0]
			    c.r, c.g, c.b, c.a = r, g, b, a
			end,
			get = function(info)
		        local c = self.db.profile.reapplycolors[0]
			    return c.r, c.g, c.b, c.a
			end,					
		},
	}
	return options
end

function DKDiseases:OnInitialize()
    -- Load the settings
    self.db = _G.LibStub("AceDB-3.0"):New("DKDiseasesDB", defaults, "Default")

	--self:MigrateSettings()

	-- Register for profile callbacks
	self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChange")
	self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChange")
	self.db.RegisterCallback(self, "OnProfileReset", "OnProfileChange")
	
	Icon:Register("DKDiseasesLDB", Broker.obj, self.db.profile.minimap)
	LSM.RegisterCallback(DKDiseases, "LibSharedMedia_Registered")
end

function DKDiseases:OnProfileChange()
	--self:MigrateSettings()
	self:Reset()
end

function DKDiseases:Reset()
	--self:ResetFonts()
	--self:UpdateTextures()

	for k, v in pairs(self.trackers) do
		if v then
			v.db = self.db.profile.trackers[v.name]
			v:Reset()
		end
	end
end

function DKDiseases:LibSharedMedia_Registered(event, mediatype, key)
	if _G.strlen(self.db.profile.font_face) > 1 and mediatype == "font" then
		if self.db.profile.font_face == key then
			self:ResetFonts()
		end
	end
	if mediatype == "statusbar" then
	    self:UpdateTextures()
	end
end

function DKDiseases:ResetFonts()
	for k, v in pairs(self.trackers) do
		if v then
			v:ResetFonts()
		end
	end
	for k, v in pairs(frames) do
		if v and v.ResetFonts then
			v:ResetFonts()
		end
	end
end

function DKDiseases:UpdateTextures()
	for k, v in pairs(self.trackers) do
		if v then
			v:UpdateTexture()
		end
	end
end

local function splitWords(str)
  local w = {}
  local function helper(word) table.insert(w, word) return nil end
  str:gsub("(%w+)", helper)
  return w
end

local function GetNumericValue(...)
    for i = 1, _G.select("#", ...) do
        local region = _G.select(i, ...)
        if region and region:GetObjectType() == "FontString" then
            local regionText = region:GetText()
            if regionText then
                local valueText = regionText:match("[0-9,]+")
                if valueText then
					valueText = valueText:gsub(",","")
                    local value = tonumber(valueText)
                    if value then
                        return value
                    end
                end
            end
        end
    end
end

local TipFrame = nil
function DKDiseases:VerifyDiseaseDamage()
	if not TipFrame then
		TipFrame = _G.CreateFrame("GameTooltip", "DKDiseases_TipFrame", nil, "GameTooltipTemplate")
		TipFrame:SetOwner(_G.WorldFrame, "ANCHOR_NONE")
	end

	TipFrame:ClearLines()
	TipFrame:SetSpellByID(SpellIds["Frost Fever"])
	local baseFF = GetNumericValue(TipFrame:GetRegions())
	TipFrame:ClearLines()
	TipFrame:SetSpellByID(SpellIds["Blood Plague"])
	local baseBP = GetNumericValue(TipFrame:GetRegions())

	self:CalculateDiseaseDamage()
	self:Print("Tooltip vs Calculated Values:")
	local fmt = "%s: Tooltip = %d,  Calculated = %d [Effective = %d]"
	self:Print(fmt:format("Frost Fever", baseFF, 
		round(self.baseFF), round(self.effFF)))
	self:Print(fmt:format("Blood Plague", baseBP, 
		round(self.baseBP), round(self.effBP)))
end

function DKDiseases:ChatCommand(input)
    if not input or input:trim() == "" then
        self:ShowOptions()
    else
		local cmds = splitWords(input)
        if cmds[1] and cmds[1] == "debug" then
			if cmds[2] and cmds[2] == "on" then
				self.db.profile.debug = true
	            self:Print("Debugging on.  Use '/dkdiseases debug off' to disable.")
		    elseif cmds[2] and cmds[2] == "off" then
				self.db.profile.debug = false
	            self:Print("Debugging off.")
			else
				self:Print("Debugging is "..(self.db.profile.debug and "on." or "off."))
			end
        elseif cmds[1] and cmds[1] == "debugBP" then
			if cmds[2] and cmds[2] == "on" then
				self.db.profile.debugBP = true
	            self:Print("Debugging BP on.")
		    elseif cmds[2] and cmds[2] == "off" then
				self.db.profile.debugBP = false
	            self:Print("Debugging BP off.")
			else
				self:Print("Debugging BP is "..(self.db.profile.debugBP and "on." or "off."))
			end
		elseif cmds[1] and cmds[1] == "verify" then
			self:VerifyDiseaseDamage()
		end
	end
end

local MandatoryEvents = {
	"PLAYER_TALENT_UPDATE",
	"PLAYER_SPECIALIZATION_CHANGED",
	"UNIT_LEVEL",
}
function DKDiseases:OnEnable()
	for i = 1, #MandatoryEvents do
		self:RegisterEvent(MandatoryEvents[i])
	end
    self:CheckTalents()
	self:ToggleDiseaseTracking()
	LoadSpellNames()
	self:Load()
	self:CheckEnabled()

	if not self.optionsFrame then
		-- Register Options
	    local displayName = _G.GetAddOnMetadata(ADDON_NAME, "Title")
		self:GetOptions()
	    _G.LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable(displayName, options)

	    self.optionsFrame = {}
	    local ACD = _G.LibStub("AceConfigDialog-3.0")
		self.optionsFrame.Main = ACD:AddToBlizOptions(
		    displayName, displayName, nil, "core")
		--self.optionsFrame.Layout = ACD:AddToBlizOptions(
		--    displayName, L["Layout"], displayName, "layout")
		ACD:AddToBlizOptions(
		    displayName, options.args.profile.name, displayName, "profile")
	    -- Register the chat command
	    self:RegisterChatCommand("dkdiseases", "ChatCommand")
	end
end

local WatchedEvents = {
	"PLAYER_REGEN_DISABLED",
	"PLAYER_REGEN_ENABLED",
	"PLAYER_TARGET_CHANGED",
	"UNIT_AURA",
	"COMBAT_LOG_EVENT_UNFILTERED",
	"PLAYER_ENTERING_WORLD",
	--"UNIT_PET",
}

function DKDiseases:Load()
	if self.isDK then
		self:CreateDisplay()
		if self.db.profile.specs[self.currentSpec] then
			if self.db.profile.frames.target.hideooc and not _G.UnitAffectingCombat("player") then
				frames.target:Hide()
			else
				frames.target:Show()
			end
			for i = 1, #WatchedEvents do
				self:RegisterEvent(WatchedEvents[i])
			end
		end
	end
end

function DKDiseases:Unload()
	for i = 1, #WatchedEvents do
		self:UnregisterEvent(WatchedEvents[i])
	end
	self:ToggleDiseaseTracking(false)
	if frames.target then frames.target:Hide() end
end

function DKDiseases:OnDisable()
	self:Unload()
end

local UnitEvents = {
	["player"] = {"UNIT_ATTACK_POWER", "UNIT_STATS"},
}
local EventFrame = _G.CreateFrame("Frame", ADDON_NAME.."_EventFrame")
local function EventFrame_OnEvent(frame, event, ...)
	if event == "UNIT_ATTACK_POWER" then
		DKDiseases:UNIT_ATTACK_POWER(event, ...)
	elseif event == "UNIT_STATS" then
		DKDiseases:UNIT_STATS(event, ...)
	end
end
EventFrame:SetScript("OnEvent", EventFrame_OnEvent)

local DiseaseStrengthEvents = {
	"PLAYER_DAMAGE_DONE_MODS",
	"COMBAT_RATING_UPDATE",
}
function DKDiseases:ToggleDiseaseTracking(flag)
	local enable = flag ~= nil and flag or self.db.profile.trackDiseases
	if self.isDK and enable then
		for i = 1, #DiseaseStrengthEvents do
			self:RegisterEvent(DiseaseStrengthEvents[i])
		end
		for unit, events in pairs(UnitEvents) do
			for i, event in ipairs(events) do
				EventFrame:RegisterUnitEvent(event, unit)
			end
		end
		self:UpdateAllStats()
	else
		for i = 1, #DiseaseStrengthEvents do
			self:UnregisterEvent(DiseaseStrengthEvents[i])
		end
		for unit, events in pairs(UnitEvents) do
			for i, event in ipairs(events) do
				EventFrame:UnregisterEvent(event)
			end
		end
	end
end

function DKDiseases:CheckEnabled()
	if self.isDK and self.db.profile.specs[self.currentSpec] then
		self:Load()
	else
		self:Unload()
	end
end

function DKDiseases:CheckTalents()
    local class, className = _G.UnitClass("player")
    if className then
        if (className == 'DEATH KNIGHT' or className == 'DEATHKNIGHT') then
            self.isDK = true

			local activeSpecNum = _G.GetSpecialization()
			if activeSpecNum and activeSpecNum > 0 then
				local id, name, desc, texture = _G.GetSpecializationInfo(activeSpecNum)
		    	if texture == "Interface\\Icons\\Spell_Deathknight_BloodPresence" then
					self.currentSpec = "Blood"
				elseif texture == "Interface\\Icons\\Spell_Deathknight_FrostPresence" then
					self.currentSpec = "Frost"
				elseif texture == "Interface\\Icons\\Spell_Deathknight_UnholyPresence" then
					self.currentSpec = "Unholy"
				else
					self:Print("Error detecing player spec.")
					self.currentSpec = "Blood"
				end
			end
        else
            self.isDK = false
        end
    end
	
	self:CheckEnabled()

	if self.db.profile.debug then
		local fmt = "CheckTalents [isDK=%s,spec=%s]"
		self:Print(
			fmt:format(tostring(self.isDK), self.currentSpec))
	end
end

local function onUpdateDisease(self, elapsed)
	self.lastUpdate = (self.lastUpdate or 0) + elapsed
	self.timer = self.timer - elapsed
	if self.lastUpdate >= MIN_UPDATE_TIME then
		local parent = self.parent
		self.lastUpdate = 0
		if self.active then
			if self.timer < 0 then
				self.timer = 0
				self.active = false
			end
		
			if parent.trackerType == "bar" then
				parent.bar:SetValue(self.timer)
			end
			--self.value:SetText(tostring(round(self.timer)))
			--self.value:Show()
		else
			--self.value:SetText("0")
			--self.value:Hide()
			if parent.trackerType ~= "icon" then
				self:Hide()
			end
			self:SetScript("OnUpdate", nil)
		end
	end
end

local fontFlags = {}
function DKDiseases:GetFontSettings()
	wipe(fontFlags)
	for k, v in pairs(self.db.profile.fontFlags) do
		if v then tinsert(fontFlags, k) end
	end
	return self.db.profile.font_face, self.db.profile.font_size, tconcat(fontFlags, ",")
end

-- Define a generic class for the progress trackers
Tracker.__index = Tracker

function Tracker:Create(settings)
	--name, fontAdj, width, height, setPoint, setValuePoint
    local object = _G.setmetatable({}, Tracker)
	object.name = settings.name
	object.parentFrame = settings.parentFrame or _G.UIParent
	object.fontAdj = settings.fontAdj or 0
	object.db = DKDiseases.db.profile.trackers[object.name]
	object.trackerType = settings.trackerType or "icon"
	object.iconTexture = settings.iconTexture
	
	local widthFunc = settings.width
	if _G.type(widthFunc) ~= "function" then
		widthFunc = function() return settings.width end
	end
	object.GetWidth = widthFunc
	local heightFunc = settings.height
	if _G.type(heightFunc) ~= "function" then
		heightFunc = function() return settings.height end
	end
	object.GetHeight = heightFunc

	object.SetPoint = settings.SetPoint
	object.SetDesiredPoint = settings.SetPoint
	object.SetValuePoint = settings.SetValuePoint
	object:Initialize()
	-- Add the bar to the addon's table of trackers
	DKDiseases.trackers[object.name] = object
	return object
end

function Tracker:Initialize()
	local tracker = _G.CreateFrame("Frame", "DKDiseases_"..self.name, self.parentFrame)
	self.tracker = tracker
	tracker.parent = self
	tracker:SetScale(1)
    tracker:SetWidth(self.GetWidth())
    tracker:SetHeight(self.GetHeight())

    local bar = _G.CreateFrame("StatusBar", "DKDiseases_"..self.name.."_Bar", tracker)
	self.bar = bar
	bar:SetAllPoints()
	self.altcolor = false
	local bt = LSM:Fetch("statusbar", DKDiseases.db.profile.texture)
    bar:SetStatusBarTexture(bt)
    bar:GetStatusBarTexture():SetHorizTile(false)
    bar:GetStatusBarTexture():SetVertTile(false)
   	local bc = self.db.color
   	bar:SetStatusBarColor(bc.r, bc.g, bc.b, bc.a)
    bar.bg = bar:CreateTexture(nil, "BACKGROUND")
    bar.bg:SetTexture(bt)
    bar.bg:SetAllPoints(true)
	local bgc = self.db.bgcolor
   	bar.bg:SetVertexColor(bgc.r, bgc.g, bgc.b, bgc.a)

    local icon = _G.CreateFrame("Frame", "DKDiseases_"..self.name.."_Icon", tracker)
	self.icon = icon
	icon:SetAllPoints()
	icon.texture = icon:CreateTexture(nil, "ARTWORK")
    icon.texture:SetAllPoints(icon)
    icon.texture:SetVertexColor(1, 1, 1)
	icon.texture:SetTexture(self.iconTexture)
	icon.cd = _G.CreateFrame("Cooldown", "DKDiseases_"..self.name.."_IconCD", icon)
	icon.cd.icon = icon
	self.iconcd = icon.cd
	icon.cd:SetAllPoints()
	
	if self.trackerType == "icon" then
		icon:Show()
		bar:Hide()
	else
		bar:Show()
		icon:Hide()
	end

	local ff, fh, fflags = DKDiseases:GetFontSettings()
	local font = LSM:Fetch("font", ff)
    tracker.value = tracker:CreateFontString(nil, "OVERLAY")
    tracker.value:SetPoint("CENTER")
    tracker.value:SetFont(font, fh + self.fontAdj, fflags)
    tracker.value:SetJustifyH("CENTER")
    --tracker.value:SetShadowOffset(1, -1)
   	local tc = self.db.textcolor
   	tracker.value:SetTextColor(tc.r, tc.g, tc.b, tc.a)
	tracker.value:Hide()

	if self.SetPoint then self:SetPoint() end
	if self.SetValuePoint then
		self:SetValuePoint()
	else
	    tracker.value:SetPoint("CENTER")
	end

	--if self.name == "FrostFever" or self.name == "BloodPlague" then
	--	tracker.active = false
	--	tracker.timer = 0
	--	tracker:Hide()
	--end

	tracker.locked = true
end

function Tracker:IsLocked()
	return self.tracker.locked
end

function Tracker:Lock(locked)
	if locked ~= nil then
		self.tracker.locked = locked
	end
	self.tracker:EnableMouse(not self.tracker.locked)
end

function Tracker:SetMovable()
	local tracker = self.tracker
    tracker:SetMovable()
    tracker:RegisterForDrag("LeftButton")
    tracker:SetScript("OnDragStart",
        function(self, button)
			if not self.locked then
            	self:StartMoving()
			end
        end)
    tracker:SetScript("OnDragStop",
        function(self)
            self:StopMovingOrSizing()
			local scale = self:GetEffectiveScale() / _G.UIParent:GetEffectiveScale()
			local x, y = self:GetCenter()
			x, y = x * scale, y * scale
			x = x - _G.GetScreenWidth()/2
			y = y - _G.GetScreenHeight()/2
			x = x / self:GetScale()
			y = y / self:GetScale()
			self.parent.db.x, self.parent.db.y = x, y
			self:SetUserPlaced(false);
        end)
	self:Lock()
end

local numberFmt = "%.0f"
function Tracker:SetValue(value)
	if self.trackerType == "bar" then
		self.bar:SetValue(value)
	else
	end
	self.tracker.value:SetText(numberFmt:format(value))
end

function Tracker:SetValueTextColor(color)
   	local tc = color or self.db.textcolor
   	self.tracker.value:SetTextColor(tc.r, tc.g, tc.b, tc.a)
end

function Tracker:Resize()
    self.tracker:SetWidth(self.GetWidth())
    self.tracker:SetHeight(self.GetHeight())
end

function Tracker:Reset()
	self:Resize()
	self:ResetFonts()
	self:UpdateVisibility()
	--self:UpdateTexture()
	self:UpdateGraphics()
end

function Tracker:ResetFonts()
	local ff, fh, fontFlags = DKDiseases:GetFontSettings()
	local font = LSM:Fetch("font", ff)
	self.tracker.value:SetFont(font, fh + self.fontAdj, fontFlags)
	self.tracker.value:SetText(self.tracker.value:GetText())
end

function Tracker:UpdateVisibility()
	if self.GetWidth then self.tracker:SetWidth(self:GetWidth()) end
	if self.GetHeight then self.tracker:SetHeight(self:GetHeight())	end
	if self.SetPoint then self:SetPoint() end
	if self.SetValuePoint then
		self:SetValuePoint()
	else
	    self.tracker.value:SetPoint("CENTER")
	end
	self:UpdateTexture()
end

function Tracker:UpdateTexture()
    local bt = LSM:Fetch("statusbar", DKDiseases.db.profile.texture)
	self.bar:SetStatusBarTexture(bt)
	self.bar.bg:SetTexture(bt)
    self.bar:GetStatusBarTexture():SetHorizTile(false)
    self.bar:GetStatusBarTexture():SetVertTile(false)
	--self:UpdateGraphics()
end

function Tracker:UpdateGraphics()
    local bc, bgc, tc
	if self.altcolor then
	    bc = self.db.alt_color or self.db.color
	    bgc = self.db.alt_bgcolor or self.db.bgcolor
	    tc = self.db.alt_textcolor or self.db.textcolor
	else
	    bc = self.db.color
	    bgc = self.db.bgcolor
	    tc = self.db.textcolor
	end

    self.bar:SetStatusBarColor(bc.r, bc.g, bc.b, bc.a)
    self.bar.bg:SetVertexColor(bgc.r, bgc.g, bgc.b, bgc.a)
    self.tracker.value:SetTextColor(tc.r, tc.g, tc.b, tc.a)
end

function Tracker:SetMinMaxValues()
end

function DKDiseases:GetBlockHeight()
	local db = DKDiseases.db.profile.frames.target
	return db.height * 2 + db.inset * 2
end

function DKDiseases:GetBlockWidth()
	local db = DKDiseases.db.profile.frames.target
	return db.width + db.inset * 2
end

function DKDiseases:CreateDiseaseBlock(name, parent)
	local baseName = parent:GetName()
	local frame = _G.CreateFrame("Frame", baseName.."_"..name.."_Block", parent)
	frame.parentFrame = parent
	frame:SetBackdrop({
		bgFile = "Interface\\Buttons\\WHITE8X8", 
		edgeFile = "",
		tile = true, tileSize = 16, edgeSize = 16,
		insets = { left = 0, right = 0, top = 0, bottom = 0 }
		})
	frame.UpdateLayout = function(self)
		self:SetWidth(DKDiseases:GetBlockWidth())
		self:SetHeight(DKDiseases:GetBlockHeight())
		local db = DKDiseases.db.profile
		local ordering = {
			["BF"] = {
				["BloodPlague"] = 1,
				["FrostFever"] = 2,
			},
			["FB"] = {
				["FrostFever"] = 1,
				["BloodPlague"] = 2,
			},
			["F"] = {
				["FrostFever"] = 1,
				["BloodPlague"] = 0,
			},
			["B"] = {
				["BloodPlague"] = 1,
				["FrostFever"] = 0,
			},
		}
		self:ClearAllPoints()
		local order = ordering[db.diseaseOrder][name]
		if order == 1 then
			self:SetPoint("BOTTOMLEFT", self.parentFrame, "BOTTOMLEFT", 0, 0)
		else
			self:SetPoint("BOTTOMRIGHT", self.parentFrame, "BOTTOMRIGHT", 0, 0)
		end
		if order == 0 then
			self:Hide()
		else
			self:Show()
		end
	end
	frame.UpdateColor = function(self, color)
		local db = DKDiseases.db.profile
		local c = color or db.reapplycolors[0]
		self:SetBackdropColor(c.r, c.g, c.b, c.a)
	end
	frame.ResetFonts = function(self)
		local ff, fh, fflags = DKDiseases:GetFontSettings()
		local font = LSM:Fetch("font", ff)
	    self.value:SetFont(font, fh, fflags)
		self.value:SetText(self.value:GetText())
	end

	frame:UpdateLayout()
	frame:UpdateColor()

    frame.value = frame:CreateFontString(nil, "OVERLAY")
    frame.value:SetPoint("TOP", frame, "TOP", 0, -self.db.profile.frames.target.inset * 3)
    frame.value:SetJustifyH("CENTER")
	frame:ResetFonts()
   	local tc = self.db.profile.trackers[name].textcolor
   	frame.value:SetTextColor(tc.r, tc.g, tc.b, tc.a)
	frame.value:Hide()

	return frame
end

function DKDiseases:CreateUnitBlock(unit)
	if frames[unit] then return end
	local unitName = unit:gsub("^%l", string.upper)
	local frame = _G.CreateFrame("Frame", "DKDiseases_"..unitName, _G.UIParent)
	frames[unit] = frame
	frame.unit = unit
	frame.unitName = unitName
	frame.db = self.db.profile.frames[unit]
	frame:SetPoint("CENTER", _G.UIParent, "CENTER", frame.db.x, frame.db.y)
	frame.lock = frame.db.locked
    frame:SetMovable()
	frame.UpdateLayout = function(self)
		frame:SetWidth(DKDiseases:GetBlockWidth() * 2)
		frame:SetHeight(DKDiseases:GetBlockHeight())
	end
	frame.Resize = function(self)
		self:UpdateLayout()
		if self.ffBlock then self.ffBlock:UpdateLayout() end
		if self.bpBlock then self.bpBlock:UpdateLayout() end
		if self.ffBar then self.ffBar:Resize() end
		if self.bpBar then self.bpBar:Resize() end
	end
	frame:UpdateLayout()
	frame.ResetFonts = function(self)
		if self.ffBlock then self.ffBlock:ResetFonts() end
		if self.bpBlock then self.bpBlock:ResetFonts() end
	end

    frame:RegisterForDrag("LeftButton")
    frame:SetScript("OnDragStart",
        function(self, button)
			if not self.lock then
            	self:StartMoving()
			end
        end)
    frame:SetScript("OnDragStop",
        function(self)
            self:StopMovingOrSizing()
			local scale = self:GetEffectiveScale() / _G.UIParent:GetEffectiveScale()
			local x, y = self:GetCenter()
			x, y = x * scale, y * scale
			x = x - _G.GetScreenWidth()/2
			y = y - _G.GetScreenHeight()/2
			x = x / self:GetScale()
			y = y / self:GetScale()
			self.db.x, self.db.y = x, y
			self:SetUserPlaced(false);
        end)
    frame:EnableMouse(not frame.lock)

	frame.ffBlock = self:CreateDiseaseBlock("FrostFever", frame)
	frame.bpBlock = self:CreateDiseaseBlock("BloodPlague", frame)

	frame.ffBar = Tracker:Create(
		{
		name = unitName.."FrostFever", 
		parentFrame = frame.ffBlock,
		trackerType = "icon",
		iconTexture = SpellIcons["Frost Fever"],
		fontAdj = 2, 
		width = function() return self.db.profile.frames[unit].width end,
		height = function() return self.db.profile.frames[unit].height end,
		SetPoint = function(self)
			self.tracker:ClearAllPoints()
			local db = DKDiseases.db.profile.frames[unit]
			self.tracker:SetPoint("BOTTOMLEFT", frame.ffBlock, "BOTTOMLEFT", 
				db.inset, db.inset)
		end,
		}
	)
	frame.ffBar:SetMovable()
	frame.ffBar.reapply = nil
	frame.ffBar.tracker:Show()

	frame.bpBar = Tracker:Create(
		{
		name = unitName.."BloodPlague", 
		parentFrame = frame.bpBlock,
		trackerType = "icon",
		iconTexture = SpellIcons["Blood Plague"],
		fontAdj = 2, 
		width = function() return self.db.profile.frames[unit].width end,
		height = function() return self.db.profile.frames[unit].height end,
		SetPoint = function(self)
			self.tracker:ClearAllPoints()
			local db = DKDiseases.db.profile.frames[unit]
			self.tracker:SetPoint("BOTTOMLEFT", frame.bpBlock, "BOTTOMLEFT", 
				db.inset, db.inset)
		end,
		}
	)
	frame.bpBar:SetMovable()
	frame.bpBar.reapply = nil
	frame.bpBar.tracker:Show()

	frame:SetAlpha(frame.db.ooc_alpha)
end

function DKDiseases:CreateDisplay()
	self:CreateUnitBlock("target")
end

-- Frost Vulnerability from Razorice is dynamic.  No need to calculate it.
-- 5.1 FF: 143 + 0.138 * AP
-- 5.1 BP: 172 + 0.138 * AP
local BASEFF = 166
local BASEBP = 197
local APMOD = 0.158
function DKDiseases:CalculateDiseaseDamage()
	local baseFF = (BASEFF + APMOD * self.effectiveAP) * self.damageModifier
	local baseBP = (BASEBP + APMOD * self.effectiveAP) * self.damageModifier

	if self.currentSpec == "Frost" then
		baseFF = baseFF * (1 + (2.0 * self.mastery / 100))
	elseif self.currentSpec == "Unholy" then
		baseFF = baseFF * 1.6
		baseBP = baseBP * (1 + (2.5 * self.mastery / 100)) * 1.6
	end

	local effFF = baseFF * (1 + self.critChance)
	local effBP = baseBP * (1 + self.critChance)

	if self.cinderglacier then
		effFF = effFF * 1.2
		effBP = effBP * 1.2
	end

	self.baseFF = baseFF
	self.baseBP = baseBP
	self.effFF = effFF
	self.effBP = effBP

	if self.db.profile.debug then
		local fmt = "FF: %d / %d, BP: %d/ %d"
		self:Print(fmt:format(
			round(baseFF), round(effFF), round(baseBP), round(effBP)))
	end

	self:CheckDiseases()
end

local percFmt = "%.0f"
local reapplyFmt = "Reapply %s [%d %s %d : %d (%.1f)]"
function DKDiseases:CheckDiseases()
	if self.currentTarget and self.db.profile.trackDiseases then
		local snapshotFF = self.FF[self.currentTarget] or 0
		local thresholdType = self.db.profile.diseaseThresholdType
		local thresholdPerc = self.db.profile.diseaseThreshold
		local thresholdAmt = self.db.profile.diseaseThresholdAmt
		local reapplyFF = 0
		local ffPerc = 0
		if snapshotFF > 0 then
			reapplyFF = 1
			ffPerc = self.effFF / snapshotFF * 100
			if self.effFF > snapshotFF then
				local threshold
				if thresholdType == "Fixed" then
					threshold = snapshotFF + thresholdAmt
				else
					threshold = snapshotFF * thresholdPerc
				end
				if self.effFF >= threshold then
					reapplyFF = 2
				else
					reapplyFF = 1
				end
			elseif self.effFF < snapshotFF then
				reapplyFF = -1
			end
			if self.db.profile.debug and reapplyFF ~= 0 then
				local diff = self.effFF-snapshotFF
				self:Print(reapplyFmt:format(
					"FF", round(self.effFF), 
					reapplyFF == -1 and "<" or ">",
					round(snapshotFF),
					diff,
					diff/snapshotFF*100))
			end
		end
		local ffBar = frames.target.ffBar
		if ffBar.trackerType == "icon" then
			ffBar.parentFrame.value:SetText(percFmt:format(ffPerc))
		end
		if reapplyFF ~= ffBar.reapply then
			if ffBar.trackerType == "icon" then
				ffBar.parentFrame:UpdateColor(self.db.profile.reapplycolors[reapplyFF])
			else
				ffBar:SetValueTextColor(self.db.profile.reapplycolors[reapplyFF])
			end
			ffBar.reapply = reapplyFF
		end

		local reapplyBP = 0
		local snapshotBP = self.BP[self.currentTarget] or 0
		local bpPerc = 0
		if snapshotBP > 0 then
			reapplyBP = 1
			bpPerc = self.effBP / snapshotBP * 100
			if self.effBP > snapshotBP then
				local threshold
				if thresholdType == "Fixed" then
					threshold = snapshotBP + thresholdAmt
				else
					threshold = snapshotBP * thresholdPerc
				end
				if self.effBP >= threshold then
					reapplyBP = 2
				else
					reapplyBP = 1
				end
			elseif self.effBP < snapshotBP then
				reapplyBP = -1
			end
			if self.db.profile.debug and reapplyBP ~= 0 then
				local diff = self.effBP-snapshotBP
				self:Print(reapplyFmt:format(
					"BP", round(self.effBP), 
					reapplyBP == -1 and "<" or ">",
					round(snapshotBP),
					diff,
					diff/snapshotBP*100))
			end
		end
		local bpBar = frames.target.bpBar
		if bpBar.trackerType == "icon" then
			bpBar.parentFrame.value:SetText(percFmt:format(bpPerc))
		end
		if reapplyBP ~= bpBar.reapply then
			if bpBar.trackerType == "icon" then
				bpBar.parentFrame:UpdateColor(self.db.profile.reapplycolors[reapplyBP])
			else
				bpBar:SetValueTextColor(self.db.profile.reapplycolors[reapplyBP])
			end
			bpBar.reapply = reapplyBP
		end
	end
end

function DKDiseases:UNIT_LEVEL(event, unit)
	if unit == "player" then
		self:CheckTalents()
	end
end

function DKDiseases:UNIT_ATTACK_POWER(event, unit)
	if unit == "player" then
		self:UpdateAllStats()
	end
end

function DKDiseases:UNIT_STATS(event, unit)
	if unit == "player" then
		self:UpdateAllStats()
	end
end

function DKDiseases:PLAYER_DAMAGE_DONE_MODS(event, unit)
	if unit == "player" then
		self:UpdateAllStats()
	end
end

function DKDiseases:COMBAT_RATING_UPDATE()
	self:UpdateAllStats()
end

function DKDiseases:UpdateAllStats()
	local update = false

	local base, posBuff, negBuff = _G.UnitAttackPower("player")
	local effAP = base + posBuff - negBuff
	if effAP ~= self.effectiveAP then
		self.effectiveAP = effAP
		update = true
	end

	local dmgMod = _G.select(7, _G.UnitDamage("player")) or 1.0
	if dmgMod ~= self.damageModifier then
		if self.db.profile.debugBP then
			local fmt = "Dmg Mod: %.1f"
			self:Print(fmt:format(dmgMod))
		end
		self.damageModifier = dmgMod
		update = true
	end

	local mastery = _G.GetMastery()
	if mastery ~= self.mastery then
		self.mastery = mastery
		update = true
	end

	local critChance = _G.GetCritChance() / 100
	if critChance > 1 then critChance = 1 end
	if critChance ~= self.critChance then
		self.critChance = critChance
		update = true
	end

	if update then self:CalculateDiseaseDamage() end
end

local LastBPTick = 0
function DKDiseases:PLAYER_REGEN_DISABLED()
	-- Player has entered combat.
	frames.target:SetAlpha(1)
	if self.db.profile.specs[self.currentSpec] then
		frames.target:Show()
	end
	LastBPTick = 0
end

function DKDiseases:PLAYER_REGEN_ENABLED()
	-- Player has left combat.
	frames.target:SetAlpha(self.db.profile.frames.target.ooc_alpha)
	if self.db.profile.frames.target.hideooc then
		frames.target:Hide()
	end
end

function DKDiseases:PLAYER_TARGET_CHANGED(unit)
	self.targetHostile = _G.UnitExists("target") and 
		_G.UnitCanAttack("player", "target") and
		not _G.UnitIsDead("target")
	self.currentTarget = _G.UnitGUID("target")

	self:UNIT_AURA("UNIT_AURA", "target")
	frames.target.ffBar.reapply = nil
	frames.target.bpBar.reapply = nil
	self:CheckDiseases()
end

local diseases
function DKDiseases:UNIT_AURA(event, unit)
	local name, rank, icon, count, dispelType, duration, expires, 
		caster, isStealable, shouldConsolidate, spellId, canApplyAura, 
		isBossDebuff, castByPlayer, value1, value2, value3 

	if unit == "target" then
		diseases = diseases or {
			[SpellNames["Frost Fever"]] = frames.target.ffBar,
			[SpellNames["Blood Plague"]] = frames.target.bpBar,
		}

		for spell, tracker in pairs(diseases) do
			name, rank, icon, count, dispelType, duration, expires, 
			caster, isStealable, shouldConsolidate, spellId, canApplyAura, 
			isBossDebuff, castByPlayer, value1, value2, value3 
			= _G.UnitDebuff("target", spell, nil, "PLAYER")

			if name and tracker.db.enabled then
				tracker.tracker.active = true
				local timeLeft = expires - _G.GetTime()
				tracker.tracker.timer = timeLeft
				if tracker.trackerType == "bar" then
					tracker.bar:SetMinMaxValues(0, duration)
					tracker.tracker:SetScript("OnUpdate", onUpdateDisease)
					tracker.tracker:Show()
				else
					tracker.parentFrame.value:Show()
					tracker.icon.cd:SetCooldown(expires - duration, duration)
				end
			else
				tracker.tracker.active = false 
				tracker.tracker.timer = 0
				tracker.tracker:SetScript("OnUpdate", nil)
				if tracker.trackerType == "icon" then
					tracker.parentFrame.value:Hide()
					tracker.parentFrame:UpdateColor()
					tracker.icon.cd:SetCooldown(0,0)
				else
					tracker.tracker:Hide()
				end
			end
		end
	elseif unit == "player" then
		if self.db.profile.trackDiseases then
			name, rank, icon, count, dispelType, duration, expires, 
			caster, isStealable, shouldConsolidate, spellId, canApplyAura, 
			isBossDebuff, castByPlayer, value1, value2, value3 
				= _G.UnitBuff("player", SpellNames["Cinderglacier"])
			local cinder = (name ~= nil)
			if cinder ~= self.cinderglacier then
				self:CalculateDiseaseDamage()
				self.cinderglacier = cinder
			end
		end
	end
end

function DKDiseases:PLAYER_SPECIALIZATION_CHANGED(event)
	self:CheckTalents()
end

function DKDiseases:PLAYER_TALENT_UPDATE(event)
	self:CheckTalents()
end

function DKDiseases:COMBAT_LOG_EVENT_UNFILTERED(...)
    local event, timestamp, eventtype, hideCaster, 
        srcGUID, srcName, srcFlags, srcRaidFlags, 
        destGUID, destName, destFlags, destRaidFlags, 
        param9, param10, param11, param12, param13, param14, 
        param15, param16, param17, param18, param19, param20

    event, timestamp, eventtype, hideCaster, 
    srcGUID, srcName, srcFlags, srcRaidFlags,
    destGUID, destName, destFlags, destRaidFlags,
    param9, param10, param11, param12, param13, param14, 
    param15, param16, param17, param18, param19, param20 = ...

	if srcGUID == self.playerGUID then
		if self.db.profile.debugBP and 
			eventtype == "SPELL_PERIODIC_DAMAGE" and 
			param9 == SpellIds["Blood Plague Periodic"] and
			destGUID == self.currentTarget and not param18 then
			if math.abs(param12 - LastBPTick) > 1 then
				LastBPTick = param12
				local fmt = "New BP Tick: %d"
				self:Print(fmt:format(LastBPTick))
			end
		end
		if eventtype == "SPELL_AURA_APPLIED" or eventtype == "SPELL_AURA_REFRESH" then
			if param9 == SpellIds["Frost Fever"] or param9 == SpellIds["Frost Fever Periodic"] then
				self.FF[destGUID] = self.effFF
				if destGUID == self.currentTarget then
					local ffBar = frames.target.ffBar
					if ffBar.trackerType == "icon" then
						ffBar.parentFrame.value:SetText("100")
						ffBar.parentFrame:UpdateColor(
							self.db.profile.reapplycolors[1])
					else
						ffBar:SetValueTextColor()
					end
					ffBar.reapply = nil
				end
			elseif param9 == SpellIds["Blood Plague"] or param9 == SpellIds["Blood Plague Periodic"] then
				self.BP[destGUID] = self.effBP
				if destGUID == self.currentTarget then
					local bpBar = frames.target.bpBar
					if bpBar.trackerType == "icon" then
						bpBar.parentFrame.value:SetText("100")
						bpBar.parentFrame:UpdateColor(
							self.db.profile.reapplycolors[1])
					else
						bpBar:SetValueTextColor()
					end
					bpBar.reapply = nil
				end
			end
		elseif eventtype == "SPELL_AURA_REMOVED" then
			if param9 == SpellIds["Frost Fever"] or param9 == SpellIds["Frost Fever Periodic"] then
				self.FF[destGUID] = nil
				frames.target.ffBar.reapply = nil
			elseif param9 == SpellIds["Blood Plague"] or param9 == SpellIds["Blood Plague Periodic"] then
				self.BP[destGUID] = nil
				frames.target.bpBar.reapply = nil
			end
		end
	end
end

function DKDiseases:PLAYER_ENTERING_WORLD(event)
	self.playerGUID = _G.UnitGUID("player") or ""
	self:UNIT_AURA("PLAYER_ENTERING_WORLD", "player")
end
